//! Layout calculations for the widget tree.
//!
//! Layout calculations are done in two passes:
//! - **Arrangement**: computes minimum sizes, net size policies, and records parent/child relationships between
//! nodes.
//! - **Layout**: given minimum sizes and net size policies from the arrangement step, computes
//! exact pixel values for positions and sizes.

use std::{ops::Deref, rc::Rc};

use cache::{Layer, NodeState};
use kahlo::math::{PixelBox, PixelSideOffsets};

mod arr;
mod cache;
mod calc;

pub use cache::{LayoutCache, LayoutNodeID};
pub use calc::recalculate;

pub struct CellMarker;
pub type TableCell = euclid::Point2D<usize, CellMarker>;
pub type TableSize = euclid::Size2D<usize, CellMarker>;

/// Sizing policy for a layout dimension. Defaults to no minimum or desired size and a low-weighted
/// desire to take up slack space.
#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
pub struct SizePolicy {
    /// The minimum number of pixels required along this dimension to avoid overlap and other
    /// issues.
    pub minimum: usize,
    /// The number of pixels requested to have a comfortable amount of space and avoid ugly-looking
    /// UI.
    pub desired: usize,
    /// How much to take up any remaining slack space.
    pub slack_weight: usize,
}

impl Default for SizePolicy {
    fn default() -> Self {
        Self {
            minimum: 0,
            desired: 0,
            slack_weight: 1,
        }
    }
}

impl SizePolicy {
    pub fn zero() -> Self {
        Self {
            minimum: 0,
            desired: 0,
            slack_weight: 0,
        }
    }

    pub fn expanding(weight: usize) -> Self {
        Self {
            minimum: 0,
            desired: 0,
            slack_weight: weight,
        }
    }

    pub fn fixed(size: usize) -> Self {
        Self {
            minimum: size,
            desired: size,
            slack_weight: 0,
        }
    }

    pub fn with_minimum(mut self, new_min: usize) -> Self {
        self.minimum = new_min;
        self.desired = self.desired.max(new_min);
        self
    }
    pub fn with_desired(mut self, new_desire: usize) -> Self {
        self.desired = new_desire;
        self
    }
    pub fn with_slack(mut self, new_slack: usize) -> Self {
        self.slack_weight = new_slack;
        self
    }

    pub fn max(self, rhs: SizePolicy) -> SizePolicy {
        Self {
            minimum: self.minimum.max(rhs.minimum),
            desired: self.desired.max(rhs.desired),
            slack_weight: self.slack_weight.max(rhs.slack_weight),
        }
    }

    pub fn max_preserve_slack(self, rhs: SizePolicy) -> SizePolicy {
        Self {
            minimum: self.minimum.max(rhs.minimum),
            desired: self.desired.max(rhs.desired),
            slack_weight: self.slack_weight,
        }
    }
}

pub struct SizePolicyTag;
pub type SizePolicy2D = euclid::Size2D<SizePolicy, SizePolicyTag>;

impl std::ops::Add<Self> for SizePolicy {
    type Output = Self;
    fn add(self, rhs: Self) -> Self::Output {
        Self {
            minimum: self.minimum + rhs.minimum,
            desired: self.desired + rhs.desired,
            slack_weight: self.slack_weight + rhs.slack_weight,
        }
    }
}

/// How to horizontally align a smaller node inside a larger node.
#[derive(Clone, Copy, PartialEq, Debug)]
pub enum HorizontalAlignment {
    Left,
    Centre,
    Right,
}

impl Default for HorizontalAlignment {
    fn default() -> Self {
        Self::Centre
    }
}

#[derive(Clone, Copy, PartialEq, Debug)]
pub enum VerticalAlignment {
    Top,
    Centre,
    Bottom,
}

impl Default for VerticalAlignment {
    fn default() -> Self {
        Self::Centre
    }
}

/// What sort of layout does a node represent?
#[derive(Clone, Copy, PartialEq, Debug)]
pub enum NodeBehaviour {
    /// A fixed rendering area, probably a root node.
    Fixed { area: PixelBox },
    /// An ordinary box sitting inside another node, hinting its size and receiving a final
    /// size assignment from its parent.
    Element,
    /// A node that floats above other content, anchored with a zero-size flexbox.
    Anchored,
}

#[derive(Clone)]
pub enum ChildArrangement {
    Custom(std::rc::Rc<dyn arr::ArrangementCalculator>),
    Container,
    Column,
    Row,
    Table,
}

impl std::fmt::Debug for ChildArrangement {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Custom(_) => f.write_str("ChildArrangement::Custom"),
            Self::Container => f.write_str("ChildArrangement::Container"),
            Self::Column => f.write_str("ChildArrangement::Column"),
            Self::Row => f.write_str("ChildArrangement::Row"),
            Self::Table => f.write_str("ChildArrangement::Table"),
        }
    }
}

impl Deref for ChildArrangement {
    type Target = dyn arr::ArrangementCalculator;
    fn deref(&self) -> &Self::Target {
        match self {
            Self::Custom(calc) => calc.as_ref(),
            Self::Container => &arr::ContainerArrangement,
            Self::Column => &arr::LineArrangement::Column,
            Self::Row => &arr::LineArrangement::Row,
            Self::Table => &arr::TableArrangement,
        }
    }
}

pub struct LayoutNode {
    /// Human-readable label for making layout tree dumps easier to read.
    label: Option<String>,
    /// Unique identifier for this LayoutNode
    cache_key: LayoutNodeID,
    /// Reference to the global layout item cache.
    cache: Rc<LayoutCache>,
    /// Layout behaviour, or how this node behaves with respect to its parent.
    behaviour: NodeBehaviour,
    /// Child arrangement, or how this node arranges its children.
    child_arrangement: ChildArrangement,
    /// Width policy: how does this widget take up horizontal space?
    width_policy: SizePolicy,
    /// Height policy: how does this widget take up vertical space?
    height_policy: SizePolicy,
    /// Horizontal alignment of this node inside its parent.
    halign: HorizontalAlignment,
    /// Vertical alignment of this node inside its parent.
    valign: VerticalAlignment,
    /// User-exposed margins, or spacing between the parent and the render area of this node.
    margin: PixelSideOffsets,
    /// Table row/column assignment for this node insite its parent.
    table_cell: Option<TableCell>,
}

impl std::fmt::Debug for LayoutNode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("LayoutNode")
            .field("label", &self.label)
            .field("cache_key", &self.cache_key)
            .field("behaviour", &self.behaviour)
            .field("child_arrangement", &self.child_arrangement)
            .field("width_policy", &self.width_policy)
            .field("height_policy", &self.height_policy)
            .field("margin", &self.margin)
            .finish()
    }
}

impl LayoutNode {
    pub fn new(cache: Rc<LayoutCache>) -> Self {
        let cache_key = LayoutNodeID::generate();
        cache.update_queue.borrow_mut().push(cache_key);
        Self {
            label: None,
            cache_key,
            cache,
            behaviour: NodeBehaviour::Element,
            child_arrangement: ChildArrangement::Column,
            width_policy: SizePolicy::default(),
            height_policy: SizePolicy::default(),
            halign: HorizontalAlignment::default(),
            valign: VerticalAlignment::default(),
            margin: PixelSideOffsets::new_all_same(0),
            table_cell: None,
        }
    }

    pub fn id(&self) -> LayoutNodeID {
        self.cache_key
    }

    pub fn cache(&self) -> &Rc<LayoutCache> {
        &self.cache
    }

    fn with_state<R, F: FnOnce(&mut NodeState) -> R>(&self, f: F) -> Option<R> {
        self.cache.with_state(self.cache_key, f)
    }

    pub fn relayout(&self) {
        self.cache.update_queue.borrow_mut().push(self.cache_key);
    }

    pub fn relayout_tree(&self) {
        let mut to_mark: Vec<LayoutNodeID> = vec![self.cache_key];
        while let Some(next) = to_mark.pop() {
            self.cache.with_state(next, |ns| {
                ns.needs_update = true;
                to_mark.extend(ns.children.iter());
            });
        }
    }

    pub fn render_area(&self) -> Option<PixelBox> {
        self.with_state(|ns| ns.area).flatten()
    }

    /// Checks if node needs to be rerendered, clearing the flag if so.
    pub fn render_check(&self) -> bool {
        self.with_state(|ns| {
            let ret = ns.needs_render;
            ns.needs_render = false;
            ret
        })
        .unwrap_or(false)
    }

    pub fn render_needed(&self) {
        self.cache.render_queue.borrow_mut().push(self.cache_key);
    }

    pub fn make_new_node(&self) -> Self {
        Self::new(self.cache.clone())
    }
}

/// Accessors
impl LayoutNode {
    pub fn label(&self) -> Option<&str> {
        self.label.as_deref()
    }
    pub fn set_label(&mut self, to: impl AsRef<str>) {
        self.label = Some(to.as_ref().to_string());
    }

    pub fn behaviour(&self) -> NodeBehaviour {
        self.behaviour
    }
    pub fn set_behaviour(&mut self, mode: NodeBehaviour) -> &mut Self {
        if self.behaviour != mode {
            self.behaviour = mode;
            self.relayout();
        }
        self
    }

    pub fn arrangement(&self) -> &ChildArrangement {
        &self.child_arrangement
    }
    pub fn set_arrangement(&mut self, arr: ChildArrangement) -> &mut Self {
        self.child_arrangement = arr;
        self.relayout();
        self
    }

    pub fn width_policy(&self) -> SizePolicy {
        self.width_policy
    }
    pub fn set_width_policy(&mut self, policy: SizePolicy) -> &mut Self {
        self.width_policy = policy;
        self.relayout();
        self
    }
    pub fn height_policy(&self) -> SizePolicy {
        self.height_policy
    }
    pub fn set_height_policy(&mut self, policy: SizePolicy) -> &mut Self {
        self.height_policy = policy;
        self.relayout();
        self
    }

    pub fn halign(&self) -> HorizontalAlignment {
        self.halign
    }
    pub fn set_halign(&mut self, halign: HorizontalAlignment) -> &mut Self {
        self.halign = halign;
        self
    }
    pub fn valign(&self) -> VerticalAlignment {
        self.valign
    }
    pub fn set_valign(&mut self, valign: VerticalAlignment) -> &mut Self {
        self.valign = valign;
        self
    }

    pub fn margin(&self) -> PixelSideOffsets {
        self.margin
    }
    pub fn margin_mut(&mut self) -> &mut PixelSideOffsets {
        &mut self.margin
    }
    pub fn margin_as_policy(&self) -> SizePolicy2D {
        SizePolicy2D::new(
            SizePolicy::fixed((self.margin.left + self.margin.right) as usize),
            SizePolicy::fixed((self.margin.top + self.margin.bottom) as usize),
        )
    }

    pub fn table_cell(&self) -> &Option<TableCell> {
        &self.table_cell
    }
    pub fn set_table_cell(&mut self, cell: Option<TableCell>) {
        self.table_cell = cell;
    }
}

impl Clone for LayoutNode {
    fn clone(&self) -> Self {
        Self {
            label: self.label.clone(),
            cache_key: LayoutNodeID::generate(),
            cache: self.cache.clone(),
            behaviour: self.behaviour,
            child_arrangement: self.child_arrangement.clone(),
            width_policy: self.width_policy,
            height_policy: self.height_policy,
            halign: self.halign,
            valign: self.valign,
            margin: self.margin,
            table_cell: self.table_cell,
        }
    }
}

impl Drop for LayoutNode {
    fn drop(&mut self) {
        self.cache.update_queue.borrow_mut().push(self.cache_key);
    }
}

fn dump_node_tree_helper(lna: LayoutNodeAccess, indent: usize, out: &mut String) {
    let ind = "    ".repeat(indent);
    out.push_str(ind.as_str());
    out.push_str("Node (");
    out.push_str(match lna.label() {
        Some(v) => v,
        None => "<anon>",
    });
    out.push_str(
        format!(
            ") [wpol: {}/{}/{} hpol: {}/{}/{} behaviour: {:?} upd: {:?} rend: {:?}]\n",
            lna.width_policy.minimum,
            lna.width_policy.desired,
            lna.width_policy.slack_weight,
            lna.height_policy.minimum,
            lna.height_policy.desired,
            lna.height_policy.slack_weight,
            lna.behaviour,
            lna.with_state(|v| v.needs_update),
            lna.with_state(|v| v.needs_render),
        )
        .as_str(),
    );
    out.push_str(ind.as_str());
    out.push_str("        ");
    out.push_str(format!("net policy:  {:?}\n", lna.with_state(|v| v.net_policy)).as_str());
    out.push_str(ind.as_str());
    out.push_str("        ");
    out.push_str(format!("render area: {:?}\n", lna.render_area()).as_str());

    for child in lna.child_iter() {
        dump_node_tree_helper(child, indent + 1, out);
    }
}

pub fn dump_node_tree(lna: LayoutNodeAccess, out: &mut String) {
    dump_node_tree_helper(lna, 0, out);
}

pub fn dump_tree_json<W: ?Sized + std::io::Write>(
    lna: LayoutNodeAccess,
    out: &mut std::io::BufWriter<W>,
) -> std::io::Result<()> {
    use std::io::Write;

    write!(out, "{{")?;
    write!(out, "\"label\": \"{}\" ", lna.label().unwrap_or("null"))?;
    write!(
        out,
        ",\"id\": {} ",
        <LayoutNodeID as Into<usize>>::into(lna.id())
    )?;
    write!(out, ",\"behaviour\": \"{:?}\"", lna.behaviour())?;
    write!(out, ",\"child_arrangement\": \"{:?}\"", lna.arrangement())?;
    write!(
        out,
        ",\"width_policy\": [{},{},{}]",
        lna.width_policy().minimum,
        lna.width_policy().desired,
        lna.width_policy().slack_weight
    )?;
    write!(
        out,
        ",\"height_policy\": [{},{},{}]",
        lna.height_policy().minimum,
        lna.height_policy().desired,
        lna.height_policy().slack_weight
    )?;
    write!(out, ",\"halign\": \"{:?}\"", lna.halign())?;
    write!(out, ",\"valign\": \"{:?}\"", lna.valign())?;
    write!(
        out,
        ",\"margin\": {{\"top\": {}, \"bottom\": {}, \"left\": {}, \"right\": {}}}",
        lna.margin().top,
        lna.margin().bottom,
        lna.margin().left,
        lna.margin().right,
    )?;
    match lna.table_cell() {
        None => write!(out, ",\"table_cell\": null")?,
        Some(c) => write!(out, ",\"table_cell\": [{},{}]", c.x, c.y)?,
    }
    match lna.render_area() {
        None => write!(out, ",\"render_area\": null")?,
        Some(a) => write!(
            out,
            ",\"render_area\": [[{},{}],[{},{}]]",
            a.min.x, a.min.y, a.max.x, a.max.y
        )?,
    }

    write!(out, ",\"children\": [")?;

    let mut needs_sep = false;
    for child in lna.child_iter() {
        if needs_sep {
            out.write(",".as_bytes())?;
        } else {
            needs_sep = true;
        }
        dump_tree_json(child, out)?;
    }

    write!(out, r#"]}}"#)?;

    Ok(())
}

/// Iterator implementation to iterate across children of a LayoutNodeContainer
#[derive(Clone)]
struct LayoutChildIter<'l> {
    lna: &'l LayoutNodeAccess<'l>,
    next_index: usize,
}

impl<'l> LayoutChildIter<'l> {
    fn new(lna: &'l LayoutNodeAccess) -> Self {
        Self { lna, next_index: 0 }
    }
}

impl<'l> Iterator for LayoutChildIter<'l> {
    type Item = LayoutNodeAccess<'l>;

    fn next(&mut self) -> Option<Self::Item> {
        let index = self.next_index;
        if index >= self.lna.child_count() {
            None
        } else {
            self.next_index += 1;
            self.lna.child(index)
        }
    }
}

pub trait LayoutTreeNode<Tag: 'static> {
    fn current_node(&self) -> &LayoutNode;
    fn child_count(&self) -> usize;
    fn child(&self, ndx: usize) -> Option<LayoutNodeAccess<'_>>;
}

const DYNPTR_WIDTH: usize =
    std::mem::size_of::<&'static dyn LayoutTreeNode<()>>() / std::mem::size_of::<*const ()>();
type DynPtr = [*const (); DYNPTR_WIDTH];

fn current_node_dispatch<'l, Tag: 'static>(dynptr: DynPtr) -> &'l LayoutNode {
    unsafe { std::mem::transmute::<_, &'l dyn LayoutTreeNode<Tag>>(dynptr).current_node() }
}

fn child_count_dispatch<'l, Tag: 'static>(dynptr: DynPtr) -> usize {
    unsafe { std::mem::transmute::<_, &'l dyn LayoutTreeNode<Tag>>(dynptr).child_count() }
}

fn child_dispatch<'l, Tag: 'static>(dynptr: DynPtr, ndx: usize) -> Option<LayoutNodeAccess<'l>> {
    unsafe { std::mem::transmute::<_, &'l dyn LayoutTreeNode<Tag>>(dynptr).child(ndx) }
}

#[derive(Clone, Copy)]
pub struct LayoutNodeAccess<'l> {
    current_node_func: fn(DynPtr) -> &'l LayoutNode,
    child_count_func: fn(DynPtr) -> usize,
    child_func: fn(DynPtr, ndx: usize) -> Option<LayoutNodeAccess<'l>>,
    dynptr: DynPtr,
}

impl<'l> LayoutNodeAccess<'l> {
    pub fn new<Tag: 'static>(from: &'l dyn LayoutTreeNode<Tag>) -> Self {
        assert_eq!(
            std::mem::size_of::<&dyn LayoutTreeNode<Tag>>(),
            std::mem::size_of::<DynPtr>()
        );
        assert_eq!(
            std::mem::align_of::<&dyn LayoutTreeNode<Tag>>(),
            std::mem::align_of::<DynPtr>()
        );
        Self {
            current_node_func: current_node_dispatch::<Tag>,
            child_count_func: child_count_dispatch::<Tag>,
            child_func: child_dispatch::<Tag>,
            dynptr: unsafe { std::mem::transmute(from) },
        }
    }

    pub fn current_node(&self) -> &'l LayoutNode {
        (self.current_node_func)(self.dynptr)
    }

    pub fn child_count(&self) -> usize {
        (self.child_count_func)(self.dynptr)
    }

    pub fn child(&self, ndx: usize) -> Option<LayoutNodeAccess<'l>> {
        (self.child_func)(self.dynptr, ndx)
    }

    pub fn child_iter(&self) -> impl Iterator<Item = LayoutNodeAccess> + Clone {
        LayoutChildIter::new(&self)
    }
}

impl<'l> Deref for LayoutNodeAccess<'l> {
    type Target = LayoutNode;
    fn deref(&self) -> &Self::Target {
        (self.current_node_func)(self.dynptr)
    }
}

/// Helper struct to store a leaf LayoutNode and automatically provide a LayoutTreeNode impl
pub struct LeafLayoutNode(LayoutNode);

impl LeafLayoutNode {
    pub fn new(ln: LayoutNode) -> Self {
        Self(ln)
    }

    pub fn access(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(self)
    }
}

impl From<LayoutNode> for LeafLayoutNode {
    fn from(value: LayoutNode) -> Self {
        Self::new(value)
    }
}

impl std::ops::Deref for LeafLayoutNode {
    type Target = LayoutNode;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl std::ops::DerefMut for LeafLayoutNode {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl LayoutTreeNode<()> for LeafLayoutNode {
    fn current_node(&self) -> &LayoutNode {
        &self.0
    }
    fn child(&self, _ndx: usize) -> Option<LayoutNodeAccess<'_>> {
        None
    }
    fn child_count(&self) -> usize {
        0
    }
}
