use std::rc::Rc;

use crate::{
    component::Component,
    layout::{LayoutNode, LayoutNodeAccess, LeafLayoutNode, SizePolicy},
    render::{RenderTarget, TextCache},
    ui::UIHandle,
};

use super::Widget;

pub struct Label<C: Component> {
    lnode: LeafLayoutNode,
    cache: TextCache,
    _ghost: std::marker::PhantomData<C>,
}

impl<C: Component> Label<C> {
    fn construct(
        font: Rc<fontdue::Font>,
        pxsize: f32,
        mut node: LayoutNode,
        text: Option<&str>,
    ) -> Self {
        node.set_width_policy(SizePolicy::fixed(0))
            .set_height_policy(SizePolicy::fixed(0));
        let mut cache = TextCache::new_with_font_and_size(font, pxsize);
        cache.update(text.unwrap_or(""));
        Self {
            lnode: LeafLayoutNode::new(node),
            cache,
            _ghost: Default::default(),
        }
    }

    pub fn new(uih: &UIHandle) -> Self {
        Self::construct(
            uih.theme().ui_font.clone(),
            uih.theme().ui_font_size,
            uih.new_layout_node(),
            None,
        )
    }

    pub fn new_with_text(uih: &UIHandle, text: &str) -> Self {
        Self::construct(
            uih.theme().ui_font.clone(),
            uih.theme().ui_font_size,
            uih.new_layout_node(),
            Some(text),
        )
    }

    pub fn new_with_node_and_text(
        font: Rc<fontdue::Font>,
        pxsize: f32,
        node: LayoutNode,
        text: &str,
    ) -> Self {
        Self::construct(font, pxsize, node, Some(text))
    }

    pub fn set_text(&mut self, text: &str) {
        self.cache.update(text);
    }
}

impl<C: Component> Widget<C> for Label<C> {
    fn poll(
        &mut self,
        uih: &mut UIHandle,
        _input_state: Option<&crate::input::InputState>,
    ) -> Vec<<C as Component>::Msg> {
        self.cache.update_node_size_hint(&mut *self.lnode);

        vec![]
    }

    fn layout_node(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(&self.lnode)
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        &mut self.lnode
    }

    fn render<'r, 'data: 'r>(&self, target: &mut RenderTarget<'r, 'data>) {
        target
            .with_node(&self.lnode)
            .simple_text(&self.cache, crate::render::ColourChoice::Foreground);

        /*if !self.lnode.render_check() {
            return;
        }
        let rcon = self.rendered.borrow();
        let Some(rendered) = rcon.as_ref() else {
            return;
        };
        let render_location = self.layout_node().render_area().unwrap();
        let render_location = render_location.inner_box(self.lnode.margin());

        surface.fill_region_masked(
            &rendered,
            PixelBox::from_size(rendered.size()),
            render_location.min,
            theme.foreground,
            kahlo::colour::BlendMode::SourceAlpha,
        );
        */
    }
}
