use std::cell::RefCell;

use crate::{
    input::{MouseButton, MouseCheckStatus},
    layout::{LayoutNode, LayoutNodeAccess, LeafLayoutNode, SizePolicy},
    prelude::*,
    render::{ColourChoice, RenderTarget, TextCache},
    ui::UIHandle,
};
use kahlo::{math::PixelBox, prelude::*};
use winit::keyboard::{Key, NamedKey};

pub struct TextEdit<C: Component> {
    lnode: LeafLayoutNode,
    text: TextCache,
    cursor_pos: Option<usize>,
    onchange: Option<Box<dyn Fn(&str) -> Option<C::Msg>>>,
}

impl<C: Component> TextEdit<C> {
    pub fn new(uih: &UIHandle) -> Self {
        let minheight = (uih.theme().ui_font_size * 1.5).round() as usize;
        let mut lnode = uih.new_layout_node();
        lnode
            .set_width_policy(SizePolicy::expanding(1))
            .set_height_policy(SizePolicy::fixed(minheight));

        Self {
            lnode: LeafLayoutNode::new(lnode),
            text: TextCache::new_with_font_and_size(
                uih.theme().ui_font.clone(),
                uih.theme().ui_font_size,
            ),
            cursor_pos: None,
            onchange: None,
        }
    }

    pub fn set_text(&mut self, to: &str) {
        self.text.update(to);
    }

    /*fn update_hints(&mut self, uih: &UIHandle) {
        let line = uih.theme().make_line(self.text.as_str());
        let (rendered, offset) = line.render_line();
        let sz = rendered.size();
        *self.rendered.borrow_mut() = Some((rendered, offset));
        self.lnode.render_needed();

        let wp = self.lnode.width_policy().with_minimum(sz.width as usize);
        self.lnode.set_width_policy(wp);
        let hp = self.lnode.height_policy().with_minimum(sz.height as usize);
        self.lnode.set_height_policy(hp);
    }*/
}

impl<C: Component> Widget<C> for TextEdit<C> {
    fn poll(
        &mut self,
        uih: &mut UIHandle,
        input_state: Option<&crate::input::InputState>,
    ) -> Vec<<C as Component>::Msg> {
        let Some(istate) = input_state else {
            return vec![];
        };

        if istate.is_focused(self.lnode.id()) {
            if let Some(kp) = &istate.keypress {
                match kp {
                    Key::Named(NamedKey::Backspace) => {
                        self.text.change(String::pop);
                        self.lnode.render_needed();
                        // self.update_hints(uih);
                    }
                    Key::Named(NamedKey::Delete) => {
                        self.text.change(String::pop);
                        self.lnode.render_needed();
                    }
                    Key::Named(NamedKey::Enter) => {
                        istate.clear_focus();
                    }
                    _ => {
                        if let Some(inp) = &istate.text_input {
                            self.text.change(|s| s.push_str(inp.as_str()));
                            self.lnode.render_needed();
                        }
                    }
                }
            }
        }
        if istate.mouse_status_in(self.lnode.render_area().unwrap(), MouseButton::Left)
            == MouseCheckStatus::Click
        {
            istate.focus_on(self.lnode.id());
        }

        // self.update_hints(uih);

        vec![]
    }

    fn layout_node(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(&self.lnode)
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        &mut self.lnode
    }

    fn render<'r, 'data: 'r>(&self, target: &mut RenderTarget<'r, 'data>) {
        target
            .with_node(&self.lnode)
            .fill(ColourChoice::Background)
            .simple_text(&self.text, ColourChoice::Foreground);

        /*if self.lnode.render_check() {
            let area = self.lnode.render_area().unwrap();
            target.fill_region(area, theme.background);

            let amap = self.rendered.borrow();
            let Some((amap, offset)) = amap.as_ref() else {
                return;
            };

            target.fill_region_masked(amap, PixelBox::from_size(amap.size()), area.min, theme.foreground, kahlo::colour::BlendMode::SourceAlpha);
        }*/
    }
}
